////////////////////////////////////////////////////////////////////// 
// mach2qtl/Main.cpp 
// (c) 2000-2007 Goncalo Abecasis, Yun Li
// 
// This file is distributed as part of the Goncalo source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile Goncalo.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Saturday August 11, 2007
// 
 
#include "CalculateResiduals.h"
#include "MathStats.h"
#include "Parameters.h"
#include "Pedigree.h"
#include "TraitTransformations.h"
#include <ctime>

#include <math.h>

int main(int argc, char ** argv)
{
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  String datfile, pedfile, mapfile;
  String infofile, dosefile, probfile;
  double rsqcutoff = 0;
  bool   inverseNormal = false;
  bool   dominant = false, recessive = false, additive = true, samplesize = false;

  clock_t startt, endt;
  startt = clock();

#ifndef VERSION
   printf("Mach2Qtl V1.1.3 (2013-04-25) -- QTL Association Mapping with Imputed Allele Counts\n"
#else
   printf("Mach2Qtl " VERSION " -- QTL Association Mapping with Imputed Allele Counts\n"
#endif
          "(c) 2007 Goncalo Abecasis, Yun Li\n\n");

  ParameterList pl;

BEGIN_LONG_PARAMETERS(longParameters)
   	LONG_PARAMETER_GROUP("Phenotypic Data")
      	LONG_STRINGPARAMETER("datfile", &datfile)
      	LONG_STRINGPARAMETER("pedfile", &pedfile)
   	LONG_PARAMETER_GROUP("Imputed Allele Counts")
      	LONG_STRINGPARAMETER("infofile", &infofile)
      	LONG_STRINGPARAMETER("dosefile", &dosefile)
    	LONG_STRINGPARAMETER("probfile", &probfile)
   	LONG_PARAMETER_GROUP("Analysis Options")
      	LONG_PARAMETER("useCovariates", &CalculateResiduals::useCovariates)
      	LONG_PARAMETER("quantileNormalization", &inverseNormal)
    	LONG_PARAMETER("dominant", &dominant)
    	LONG_PARAMETER("recessive", &recessive)
    	LONG_PARAMETER("additive", &additive)
   	LONG_PARAMETER_GROUP("Output")
      	LONG_PARAMETER("samplesize", &samplesize)
	LONG_DOUBLEPARAMETER("rsqcutoff", &rsqcutoff)
END_LONG_PARAMETERS();

  pl.Add(new LongParameters("Available Options", longParameters));

  pl.Add(new HiddenString('d', "Data File", datfile));
  pl.Add(new HiddenString('p', "Pedigree File", pedfile));
  pl.Add(new HiddenString('m', "Map File", mapfile));
  pl.Add(new HiddenString('g', "Dosage File", dosefile));
  pl.Add(new HiddenString('i', "Information File", infofile));

  pl.Read(argc, argv);
  pl.Status();

  // Load pedigree data
  Pedigree ped;

  ped.Prepare(datfile);
  ped.Load(pedfile);

  if (inverseNormal)
  {
    printf("Applying quantile normalization to each phenotype ...\n\n");
    InverseNormalTransform(ped);
  }

  // Adjust trait values
  CalculateResiduals baseModel;

  baseModel.FitLinearModels(ped);

  // Load marker information
  printf("\nLoading marker information ...\n");

  StringArray markers;

  markers.Read(infofile);

  for (int i = markers.Length() - 1; i >= 0; i--)
  {
    markers[i].Trim();

    if (markers[i].IsEmpty())
      markers.Delete(i);
  }

  printf("   %d markers will be analyzed\n\n", markers.Length() - 1);

  IFILE probf = ifopen(probfile, "rb");
   
  // for dominant/recessive models, need prob file
  if (dominant || recessive)
    if (probf == NULL)
    {
      printf("probfile not available, dominant/recessive models will NOT be fitted\n");
      dominant = false;
      recessive = false;
    }

  // Update test statistics gradually
  int tests = (markers.Length() - 1) * ped.traitCount;

  Vector   A(tests), D(tests), R(tests), Y(ped.traitCount);
  Vector   AA(tests), DD(tests), RR(tests);
  Vector   AY(tests), DY(tests), RY(tests), YY(ped.traitCount);
  IntArray n(ped.traitCount);

  A.Zero(); Y.Zero();
  AA.Zero(); YY.Zero(); AY.Zero();

  n.Zero();

  // use dosefile only when probfile is not available
  if (probf == NULL)
  {
    // Process dosage file
    printf("Processing dosage file ...\n\n");

    IFILE f = ifopen(dosefile, "rb");

    if (f == NULL)
    {
     	printf("dosefile CANNOT be opened \n\n");
	return(1);
    }

    String      buffer, famid, pid;
    StringArray tokens;

    printf("Executing first pass analysis of imputed genotypes ...\n");

    while (!ifeof(f))
    {
      tokens.ReplaceTokens(buffer.ReadLine(f));

      if (tokens.Length() == 0) continue;

      if (tokens.Length() != markers.Length() + 1)
      {
        printf("Expecting %d markers, but found %d markers for individuals %s\n",
          markers.Length() - 1, tokens.Length() - 2, (const char *) tokens[0]);
        continue;
      }

      int arrow = tokens[0].Find("->");

      if (arrow < 0)
      {
        printf("Labels should be FAMILY->INDIVIDUAL, cannot decode individual %s\n", (const char *) tokens[0]);
        continue;
      }

      famid = tokens[0].Left(arrow);
      pid = tokens[0].SubStr(arrow + 2);

      Person * person = ped.FindPerson(famid, pid);

      if (person == NULL)
      {
        // printf("Individual %s is not in pedigree\n", (const char *) tokens[0]);
        continue;
      }

      for (int j = 0, index = 0; j < markers.Length() - 1; j++)
      {
        double dose = tokens[j + 2].AsDouble();

        for (int i = 0; i < ped.traitCount; i++, index++)
          if (person->traits[i] != _NAN_)
            A[index] += dose;
      }

      for (int i = 0; i < ped.traitCount; i++)
        if (person->traits[i] != _NAN_)
        {
          Y[i] += person->traits[i];
          n[i] ++;
        }
    }	// End of while (first pass analysis)

    for (int j = 0, index = 0; j < markers.Length() - 1; j++)
      for (int i = 0; i < ped.traitCount; i++, index++)
        if (n[i])
          A[index] /= n[i];

    for (int i = 0; i < ped.traitCount; i++)
      if (n[i])
        Y[i] /= n[i];

    printf("Executing second pass analysis of imputed genotypes ...\n");

    ifrewind(f);

    while (!ifeof(f))
    {
      tokens.ReplaceTokens(buffer.ReadLine(f));

      if (tokens.Length() == 0) continue;

      if (tokens.Length() != markers.Length() + 1)
      {
        printf("Expecting %d markers, but found %d markers for individuals %s\n",
          markers.Length() - 1, tokens.Length() - 2, (const char *) tokens[0]);
        continue;
      }

      int arrow = tokens[0].Find("->");

      if (arrow < 0)
      {
        printf("Labels should be FAMILY->INDIVIDUAL, cannot decode individual %s\n", (const char *) tokens[0]);
        continue;
      }

      famid = tokens[0].Left(arrow);
      pid = tokens[0].SubStr(arrow + 2);

      Person * person = ped.FindPerson(famid, pid);

      if (person == NULL)
      {
        // printf("Individual %s is not in pedigree\n", (const char *) tokens[0]);
        continue;
      }

      for (int j = 0, index = 0; j < markers.Length() - 1; j++)
      {
        double dose = tokens[j + 2].AsDouble();

        for (int i = 0; i < ped.traitCount; i++, index++)
          if (person->traits[i] != _NAN_)
          {
            double adjusted_dose  = dose - A[index];
            double adjusted_trait = person->traits[i] - Y[i];

            AY[index] += adjusted_dose * adjusted_trait;
            AA[index] += adjusted_dose * adjusted_dose;

            // printf("AY = %.5f, AA = %.5f\n", AY[index], AA[index]);
          }
      } 

      for (int i = 0; i < ped.traitCount; i++)
        if (person->traits[i] != _NAN_)
        {
          double adjusted_trait = person->traits[i] - Y[i];
          YY[i] += adjusted_trait * adjusted_trait;
        }
    }	// End of while (second pass)

    ifclose(f);
    // end of Process dosage file
  }

  // when probfile available
  else 
  {
    // Process prob file
    printf("Processing prob file ... \n\n");

    String      buffer, famid, pid;
    StringArray tokens;

    printf("Executing first [ass analysis of imputed genotypes ... \n");

    while (!ifeof(probf))
    {
      tokens.ReplaceTokens(buffer.ReadLine(probf));

      if (tokens.Length() == 0) continue;

      if (tokens.Length() != (markers.Length() * 2) )
      {
        printf("Expecting %d markers, but found %d markers for individual %s\n",
                markers.Length() - 1, tokens.Length() / 2 - 1, (const char *) tokens[0]);
        continue;
      }

      int arrow = tokens[0].Find("->");

      if (arrow < 0)
      {
        printf("Labels should be FAMILY->INDIVIDUAL, cannot decode individual %s\n", 
               (const char *) tokens[0]);
        continue;
      }

      famid = tokens[0].Left(arrow);
      pid   = tokens[0].SubStr(arrow+2);

      Person * person = ped.FindPerson(famid, pid);

      if (person == NULL)
      {
        // printf ("Individual %s is not in pedigree\n", (const char *) tokens[0]);
        continue;
      }

      for (int j = 0, index = 0; j < markers.Length() - 1; j++)
      {
        double pAA = tokens[2*j + 2].AsDouble();
        double pAB = tokens[2*j + 3].AsDouble();
        double pBB = 1.00 - pAA - pAB;

        double dose = pAA * 2 + pAB;
        double Aprob = pAA + pAB;
        double Bprob = pAB + pBB;

        for (int i = 0; i < ped.traitCount; i++, index++)
          if (person->traits[i] != _NAN_)
          {
            A[index] += dose;
            D[index] += Aprob;
            R[index] += Bprob;
          }
      }

      for (int i = 0; i < ped.traitCount; i++)
        if (person->traits[i] != _NAN_)
        {
          Y[i] += person->traits[i];
          n[i] ++;
        }

    }// end of while (first pass)

    for (int j = 0, index = 0; j < markers.Length() - 1; j++)
      for (int i = 0; i < ped.traitCount; i++, index ++)
        if (n[i])
        {
          A[index] /= n[i];
          D[index] /= n[i];
          R[index] /= n[i];
        }

    for (int i = 0; i < ped.traitCount; i++)
      if (n[i])
        Y[i] /= n[i];

    printf("Executing second pass analysis of imputed genotypes ... \n");

    ifrewind(probf);

    while (!ifeof(probf))
    {
      tokens.ReplaceTokens(buffer.ReadLine(probf));

      if (tokens.Length() == 0) continue;

      if (tokens.Length() != (markers.Length() * 2) )
      {
        printf("Expecting %d markers, but found %d markers for individual %s\n",
               markers.Length() - 1, tokens.Length() / 2 - 1, (const char *) tokens[0]);
        continue;
      }
       
      int arrow = tokens[0].Find("->");

      if (arrow < 0)
      {
        printf("Labels should be FAMILY->INDIVIDUAL, cannot decode individual %s\n", 
              (const char *) tokens[0]);
        continue;
      }

      famid = tokens[0].Left(arrow);
      pid = tokens[0].SubStr(arrow + 2);

      Person * person = ped.FindPerson(famid, pid);

      if (person == NULL)
      {
        // printf("Individual %s is not in pedigree\n", (const char *) tokens[0]);
        continue;
      }

      for (int j = 0, index = 0; j < markers.Length() - 1; j++)
      {
        double pAA = tokens[2*j + 2].AsDouble();
        double pAB = tokens[2*j + 3].AsDouble();
        double pBB = 1.00 - pAA - pAB;

        double dose = pAA * 2 + pAB;
        double Aprob = pAA + pAB;
        double Bprob = pAB + pBB;

        for (int i = 0; i < ped.traitCount; i++, index++)
          if (person->traits[i] != _NAN_)
          {
            double adjusted_dose = dose - A[index];
            double adjusted_Aprob = Aprob - D[index];
            double adjusted_Bprob = Bprob - R[index];
            double adjusted_trait = person->traits[i] - Y[i];

            AY[index] += adjusted_dose * adjusted_trait;
            DY[index] += adjusted_Aprob * adjusted_trait;
            RY[index] += adjusted_Bprob * adjusted_trait;

            AA[index] += adjusted_dose * adjusted_dose;
            DD[index] += adjusted_Aprob * adjusted_Aprob;
            RR[index] += adjusted_Bprob * adjusted_Bprob;
          }
      }

      for (int i = 0; i < ped.traitCount; i++)
        if (person->traits[i] != _NAN_)
        {
          double adjusted_trait = person->traits[i] - Y[i];
          YY[i] += adjusted_trait * adjusted_trait;
        }
       
    } // end of while (!ifeof(probf))

    ifclose(probf);
  } // end of else: when probfile available

  printf("\n");
  printf("                \tINFORMATION FROM .info FILE          ");
  if (additive)
    printf("   QTL ASSOCIATION ADDITIVE         ");
  if (dominant)
    printf("   QTL ASSOCIATION DOMINANT         ");
  if (recessive)
    printf("   QTL ASSOCIATION RECESSIVE        ");
  printf("\n");

  printf("                 \t=====================================");
  if (additive)
    printf("   =================================");
  if (dominant)
    printf("   =================================");
  if (recessive)
    printf("   =================================");
  printf("\n");

  printf("TRAIT            \tMARKER               ALLELES  FREQ1  RSQR");
  if (additive)
    printf("   EFFECT2  STDERR  CHISQ     PVALUE");
  if (dominant)
    printf("   EFFECT   STDERR  CHISQ     PVALUE");
  if (recessive)
    printf("   EFFECT   STDERR  CHISQ     PVALUE");
  if (samplesize)
    printf(" N");
  printf("\n");


  for (int j = 0, index = 0; j < markers.Length() - 1; j++)
    for (int i = 0; i < ped.traitCount; i++, index++)
      if (n[i] >= 2)
      {
        StringArray tokens;
        tokens.ReplaceTokens(markers[j + 1]);

// If rsq equals 0 (default), output results for all the markers.
// Skip outputing markers if given an rsq threshold.
        if (tokens[6].AsDouble() > 0)
	{
		if (tokens[6].AsDouble() <= rsqcutoff)
        	{
                	continue;
        	}
	}

        printf("%-17s\t%-21s", (const char *) ped.traitNames[i], (const char *) tokens[0]);

        if (tokens.Length() < 5)
          printf("** Marker Info Error **  ");
        else
          printf("%-9s", (const char *) (tokens[1] + "," + tokens[2]));

        if (tokens[3].AsDouble() >= 0.99995)
          printf("1.000   ", tokens[3].AsDouble());
        else
          printf(".%04.0f  ", tokens[3].AsDouble() * 10000.0);

        if (tokens[6].AsDouble() >= 0.99995)
          printf("1.000   ", tokens[6].AsDouble());
        else
          printf(".%04.0f  ", tokens[6].AsDouble() * 10000.0);

        double variance    = YY[i] / (n[i] - 1);

        if (variance <= 1e-10) continue;

        if (additive)
        {
          double numerator   = AY[index] / variance;
          double denominator = AA[index] / variance + 1e-30;

          double statistic = numerator * numerator / denominator;
          double effect = -numerator / denominator;
          double serror = 1.0 / sqrt(denominator);

          int digits = (effect < 999.9995 && effect > -99.9995) ? 3 :
                 (effect < 9999.995 && effect > -999.995) ? 2 :
                 (effect < 99999.95 && effect > -9999.95) ? 1 : 0;

          int sdigits = (serror < 999.9995 && serror > -99.9995) ? 3 :
                  (serror < 9999.995 && serror > -999.995) ? 2 :
                  (serror < 99999.95 && serror > -9999.95) ? 1 : 0;

	  //digits += 2;
          //sdigits += 2;
          //printf("% -8.*f %-7.*f %-9.5f %8.5g   ",
          printf("% -8.*f %-7.*f %-8.4f %7.4g   ",
              digits, effect, sdigits, serror,
              statistic, statistic > 0 ? chidist(statistic,1) : 1.0);
        }
        if (dominant)
        {
          double numerator   = DY[index] / variance;
          double denominator = DD[index] / variance + 1e-30;

          double statistic = numerator * numerator / denominator;
          double effect = -numerator / denominator;
          double serror = 1.0 / sqrt(denominator);


          int digits = (effect < 999.9995 && effect > -99.9995) ? 3 :
                 (effect < 9999.995 && effect > -999.995) ? 2 :
                 (effect < 99999.95 && effect > -9999.95) ? 1 : 0;

          int sdigits = (serror < 999.9995 && serror > -99.9995) ? 3 :
                  (serror < 9999.995 && serror > -999.995) ? 2 :
                  (serror < 99999.95 && serror > -9999.95) ? 1 : 0;

          printf("% -8.*f %-7.*f %-8.4f %7.4g   ",
              digits, effect, sdigits, serror,
              statistic, statistic > 0 ? chidist(statistic,1) : 1.0);
        }
        if (recessive)
        {
          double numerator   = RY[index] / variance;
          double denominator = RR[index] / variance + 1e-30;

          double statistic = numerator * numerator / denominator;
          double effect = -numerator / denominator;
          double serror = 1.0 / sqrt(denominator);


          int digits = (effect < 999.9995 && effect > -99.9995) ? 3 :
                 (effect < 9999.995 && effect > -999.995) ? 2 :
                 (effect < 99999.95 && effect > -9999.95) ? 1 : 0;

          int sdigits = (serror < 999.9995 && serror > -99.9995) ? 3 :
                  (serror < 9999.995 && serror > -999.995) ? 2 :
                  (serror < 99999.95 && serror > -9999.95) ? 1 : 0;

          printf("% -8.*f %-7.*f %-8.4f %7.4g   ",
              digits, effect, sdigits, serror,
              statistic, statistic > 0 ? chidist(statistic,1) : 1.0);
        }

        if (samplesize) printf(" %d", n[i]);
        printf("\n");
      
      }
      
      
  endt = clock();
  int lapsetime = (int) ((double)(endt - startt) / CLOCKS_PER_SEC);
  printf("\nAnalysis took %d seconds\n\n", lapsetime);

}
 
